#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Régression sans inversion
# 
# Ce notebook mesure le temps de calcul dans deux algorithmes pour résoudre une régression linéaire, le premier inverse un matrice, le second le fait sans inverser une matrice, le troisième reprend l'idée du second mais utilise une décomposition [QR](https://fr.wikipedia.org/wiki/D%C3%A9composition_QR) puis inverse la matrice *R*.

# In[1]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[2]:


import numpy.random as rnd
X = rnd.randn(1000, 7)
eps = rnd.randn(1000, 1) / 3
y = X.sum(axis=1).reshape((X.shape[0], 1)) + eps
y = y.ravel()
X.shape, y.shape, eps.shape


# In[3]:


from mlstatpy.ml.matrices import linear_regression, gram_schmidt
beta1 = linear_regression(X, y, algo=None)
beta2 = linear_regression(X, y, algo="gram")
beta1, beta2


# In[4]:


get_ipython().run_line_magic('timeit', 'linear_regression(X, y, algo=None)')


# In[5]:


get_ipython().run_line_magic('timeit', 'linear_regression(X, y, algo="gram")')


# In[6]:


get_ipython().run_line_magic('timeit', 'linear_regression(X, y, algo="qr")')


# In[7]:


Xt = X.T
get_ipython().run_line_magic('timeit', 'gram_schmidt(Xt)')


# Un exemple avec [scikit-learn](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html).

# In[8]:


from sklearn.linear_model import LinearRegression
clr = LinearRegression()
get_ipython().run_line_magic('timeit', 'clr.fit(X, y)')


# Qui utilise la fonction [lstsq](https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.lstsq.html?highlight=lstsq):

# In[9]:


from numpy.linalg import lstsq
get_ipython().run_line_magic('timeit', 'lstsq(X, y, rcond=None)')


# Il serait sans doute possible d'optimiser les calculs en réduisant le nombre de copie et de transposées. La version utilisant une décomposition [QR](https://fr.wikipedia.org/wiki/D%C3%A9composition_QR) est assez rapide. Le code est là [matrices.py](https://github.com/sdpython/mlstatpy/blob/master/src/mlstatpy/ml/matrices.py). Pour dépasser [numpy](https://www.numpy.org/), il faut passer au C++. *scikit-learn* ajoute des étapes intermédiaires pour vérifier les données ce qui explique la longueur. On résume le tout par un graphique.

# In[10]:


from cpyquickhelper.numbers import measure_time


# In[11]:


stmts = [dict(name='lr_matrix', fct="linear_regression(X, y, algo=None)"),
         dict(name='lr_gram', fct="linear_regression(X, y, algo='gram')"),
         dict(name='lr_qr', fct="linear_regression(X, y, algo='qr')"),
         dict(name='gram', fct="gram_schmidt(Xt)"),
         dict(name='sklearn', fct="clr.fit(X, y)"),
         dict(name='lstsq', fct="lstsq(X, y)")]

memo = []
for size, dim in [(100, 10), (1000, 10), (10000, 10),
                  (100, 20), (1000, 20), (10000, 20),
                  (100, 50), (1000, 50)]:
    print(size, dim)
    X = rnd.randn(size, dim)
    eps = rnd.randn(size, 1) / 3
    y = X.sum(axis=1).reshape((X.shape[0], 1)) + eps
    y = y.ravel()
    context = dict(linear_regression=linear_regression, Xt=X.T,
                   X=X, y=y, gram_schmidt=gram_schmidt, clr=clr,
                   lstsq=lambda X, y: lstsq(X, y, rcond=None))
    
    for stmt in stmts:
        res = measure_time(stmt['fct'], number=20, repeat=20, div_by_number=True, context=context)
        res.update(stmt)
        res['size'] = size
        res['dim'] = dim
        memo.append(res)

import pandas
df = pandas.DataFrame(memo)
df.head()


# In[12]:


piv = pandas.pivot_table(df, index=['size', 'dim'], columns='name', values='average')
piv


# In[13]:


import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 2, figsize=(14,4))
piv[:6].plot(kind="bar", ax=ax[0])
piv[6:].plot(kind="bar", ax=ax[1])
ax[0].set_title("Régression Linéaire, size < 10000")
ax[1].set_title("Régression Linéaire, size >= 10000");


# ## Streaming versions
# 
# L'idée est différente ici puisqu'il s'agit de calculer toutes les régressions linéaires intermédiaires. Les algorithmes sont décrits par l'exposé [Régression linéaire par morceaux](http://www.xavierdupre.fr/app/mlstatpy/helpsphinx/c_ml/piecewise.html#streaming-linear-regression).

# In[14]:


from mlstatpy.ml.matrices import streaming_linear_regression, streaming_linear_regression_gram_schmidt

def all_linear_regression(X, y):
    for i in range(X.shape[1], X.shape[0]):
        yield linear_regression(X[:i], y[:i])

stmts = [dict(name='lr_matrix', fct="list(all_linear_regression(X, y))"),
         dict(name='lr_st_mat', fct="list(streaming_linear_regression(X, y))"),
         dict(name='lr_st_gram', fct="list(streaming_linear_regression_gram_schmidt(X, y))"),
        ]

memo = []
for dim in (10, ):
    for size in range(100, 3500, 500):
        print(size, dim)
        X = rnd.randn(size, dim)
        eps = rnd.randn(size, 1) / 3
        y = X.sum(axis=1).reshape((X.shape[0], 1)) + eps
        y = y.ravel()
        context = dict(X=X, y=y,
                       all_linear_regression=all_linear_regression,
                       streaming_linear_regression=streaming_linear_regression,
                       streaming_linear_regression_gram_schmidt=streaming_linear_regression_gram_schmidt)

        for stmt in stmts:
            if "gram" in stmt['name']:
                nn = 1
                if size >= 1000:
                    continue
            else:
                nn = 5
            res = measure_time(stmt['fct'], number=nn, repeat=nn, div_by_number=True, context=context)
            res.update(stmt)
            res['size'] = size
            res['dim'] = dim
            memo.append(res)

import pandas
df = pandas.DataFrame(memo)
df.head()


# In[15]:


piv = pandas.pivot_table(df, index=['size'], columns='name', values='average')
piv


# In[16]:


fig, ax = plt.subplots(1, 2, figsize=(14,4))
piv[["lr_matrix", "lr_st_mat"]].plot(ax=ax[0])
piv.plot(ax=ax[1])
ax[0].set_title("Régression Linéaire streaming (all)\n10 features")
ax[1].set_title("Régression Linéaire no Gram-Schmidt\n10 features");


# La version streaming devient plus intéressante à partir de 1000 observations, le coût en linéaire en *N* contrairement à la version classique qui est en $N^2$. La version Gram-Schmidt devrait être réécrite en C++ pour proposer des temps comparables.

# In[17]: